from enum import IntEnum


class DeviceType(IntEnum):
    OMEGA33 = 34
    OMEGA33_LEFT = 36
    OMEGA331 = 35
    OMEGA331_LEFT = 37
    SIGMA331 = 104
    SIGMA331_LEFT = 105
    LAMBDA331 = 108
    LAMBDA331_LEFT = 109
